from typing import List

from pygame.event import Event
from pygame.surface import Surface
from pygame.locals import *
from pygame import draw

from Controls import Controls
from CarouselItem import CarouselItem
from CarouselIndicators import CarouselIndicators
import Colors
from Manifest import Manifest
from Controls import Controls
from ScrollBar import *

class Carousel:
    """Creates a Carousel Component with the given elements
       Occupies more or less the entire surface size. 
       The currently active element is displayed in the center.
       While the prev and next elements are also displayed.
       Also has a Carousel Indicator on top. 
    """
    def __init__(self,screen: Surface,originX:float,originY:float, width:float,height:float,identifier:str,manifestList:List,onChange=None,onChanged=None):
        """ Initialise the Carousel

        Args:
            surface (Surface): The surface/surface to display the carousel on.
            originX (float): x position of the top left corner
            originY (float): y position of the top left corner
            width (float): Width of the Carousel
            height (float): Height of the Carousel
            identifier (str): String name for identifying the carousel
            manifestList (List): List of elements to display in the carousel
            onChange (_type_, optional): Callback for onChange event. Fired at the start of a slide change. Defaults to None.
            onChanged (_type_, optional):Callback for onChanged event. Fired after the slide is changed. Defaults to None.
        """
        self.screen = screen
        self.surface = Surface((width,height))
        self.x = originX
        self.y = originY
        self.width = width
        self.height = height
        self.identifier = identifier



        # Parameters to help with animation and slide change
        self.isAnimating = False
        # Count the current number of frames.
        self.count = 0
        # keep track of the amount of to move the rects by.
        self.newPos = [0,0]
        self.total_animation_frames = 10
        self.spacing = 20
        # Set the callbacks for the events
        self.onChange = onChange
        self.onChanged = onChanged
       
        # Set the carousel elements list
        self.manifestList = manifestList

        # Intialise the carousel Items and the active index
        self.carouselItemList: List[CarouselItem] = []
        self.activeIndex = 0

        # Initialise the Surfaces and rects of the Carousel Items
        for x,manifest in enumerate(self.manifestList):
            item = CarouselItem(self.surface,manifest)
            item.rect.centery = int(height /2)
            if x == 0:
                item.rect.centerx = int(width/2)
            elif x == (len(self.manifestList) - 1):
                item.rect.right = self.carouselItemList[0].rect.left - self.spacing
            else:
                item.rect.left = self.carouselItemList[len(self.carouselItemList)-1].rect.right + self.spacing
            self.carouselItemList.append(item)
        

        slice_x = width/12
        slice_y = height/12

        scrolly = self.carouselItemList[self.activeIndex].rect.y/4
        default_width = 2*slice_x
        self.scrollBar = ScrollBar(self.surface, scrolly, default_width, len(self.carouselItemList), self.total_animation_frames)


        # Focus on the first item
        self.focus()

    def getActiveIndex(self):
        return self.activeIndex

    def setActiveIndex(self,newIndex):
        self.carouselItemList[self.activeIndex].blur()
        change = self.surface.get_rect().centerx - self.carouselItemList[newIndex].rect.centerx
        
        if change == 0:
            self.focus()
            return
        
        self.newPos = [ change, 0]


        self.activeIndex = newIndex
        if change > 0: 
            self.carouselItemList[(newIndex - 1)% len(self.carouselItemList)].rect.right = self.carouselItemList[newIndex].rect.left - self.spacing
        elif change < 0:
            self.carouselItemList[(newIndex + 1)% len(self.carouselItemList)].rect.left = self.carouselItemList[newIndex].rect.right + self.spacing

        
        print('setIndex change',change)
        
        # Calculate slider change
        sliderChange = self.scrollBar.scrollBarSurface.get_width()/len(self.carouselItemList)*self.activeIndex
        
        # Change the slider position 
        self.scrollBar.sliderRect.left = self.scrollBar.scrollBar.left + sliderChange

        # Change the rect positions 
        for x,carouselItem in enumerate(self.carouselItemList):
            carouselItem.rect = carouselItem.rect.move(self.newPos)
        

        self.newPos = [0,0]
        self.focus()

    def carouselMoveAnimation(self,newPosx,newPosy):
        # Calculate speed so that movement happens in even steps 
        speed = [int(newPosx/self.total_animation_frames),int(newPosy/self.total_animation_frames)]    
        # Move items accoring to speed for every animation frame
        if self.count == self.total_animation_frames:
            speed = [newPosx - speed[0]*(self.total_animation_frames - 1)  ,newPosy  - speed[1]*(self.total_animation_frames - 1)]
        

        if self.count <= self.total_animation_frames:
            for x,carouselItem in enumerate(self.carouselItemList):
                self.surface.blit(self.screen,carouselItem.rect,carouselItem.rect)
                carouselItem.rect = carouselItem.rect.move(speed)
                self.surface.blit(carouselItem.surface,carouselItem.rect)
   
            # Move slider concurrently
            self.scrollBar.animateSlider(newPosx, self.activeIndex)

        else:
            self.count = 0
            self.isAnimating = False
            self.newPos = [0,0]
            self.focus()
            self.handleOnChanged()

    def prev(self):
        newIndex = (self.activeIndex - 1) % len(self.carouselItemList)
        newIndex %= len(self.carouselItemList)
        if not self.isAnimating:
            # Deactivate the active item
            self.carouselItemList[self.activeIndex].blur()

            change = self.surface.get_rect().centerx - self.carouselItemList[newIndex].rect.centerx 
            
            self.newPos = [ change, 0]

            self.activeIndex = newIndex
            self.carouselItemList[(newIndex - 1)% len(self.carouselItemList)].rect.right = self.carouselItemList[newIndex].rect.left - self.spacing
            self.isAnimating = True
            self.handleOnChange()
        
    def next(self):
        newIndex = (self.activeIndex + 1) % len(self.carouselItemList)
        newIndex %= len(self.carouselItemList)

        if not self.isAnimating:
            # Deactivate the active item
            self.carouselItemList[self.activeIndex].blur()

            change = self.surface.get_rect().centerx - self.carouselItemList[newIndex].rect.centerx
            
            self.newPos = [ change, 0]

            self.activeIndex = newIndex
            self.carouselItemList[(newIndex + 1)% len(self.carouselItemList)].rect.left = self.carouselItemList[newIndex].rect.right + self.spacing
            self.isAnimating = True
            self.handleOnChange()     

    def focus(self):
        index = self.activeIndex
        self.surface.blit(self.screen,self.carouselItemList[index].rect,self.carouselItemList[index].rect)            
        self.carouselItemList[index].focus()


    def onSelect(self,callback=None,cargs=()):
        print('Carousel onSelect')
        if callback is not None:
            self.carouselItemList[self.activeIndex].onSelect(callback,cargs)
    
    def handleOnChange(self):
        if self.onChange is not None:
            self.onChange(self)

    def handleOnChanged(self):
        if self.onChanged is not None:
            self.onChanged(self)
    
    def moveCursor(self,event: Event,controls: Controls):
        if controls.is_movement_left(event):
            self.prev()
        elif controls.is_movement_right(event):
            self.next()
        
    
    def render(self):
        self.surface.fill(Colors.BLACK)
        if self.isAnimating:
            self.count+= 1
            self.carouselMoveAnimation(self.newPos[0],self.newPos[1])

        # Draw all the remaining items 
        for x in range(len(self.carouselItemList)):
            if ( x != self.activeIndex):
                self.surface.blit(self.carouselItemList[x].surface,self.carouselItemList[x].rect)
        
        # Draw the active Item last so that it draws over the other items
        self.surface.blit(self.carouselItemList[self.activeIndex].surface, self.carouselItemList[self.activeIndex].rect)

        # Draw The Scroll Bar
        self.scrollBar.render()
        
        # Render the Carousel to Screen
        self.screen.blit(self.surface,self.surface.get_rect())


        self.screen.blit(self.surface, self.surface.get_rect())


        
