"""
Controls.py - Controls abstraction classes for the fabarcade.
Created by MCP Group 2 
Authors: Shailesh Iyer, Sijia Guo,Lina Gönnheimer,Katharina Güths
=========== DISCLAIMER ==============================================================
Provided as is. May contain bugs with your game so make sure to test on the fabarcade.
Also you may have to do input debouncing for chorded input.  
I have tested the latest version on the pi and the fabarcade.It works decently well.
=====================================================================================
Control mapping for the fabArcade
    ------------------------------------------------------   
    | FabArcade Button  | XBox Button   | Pygame Number  |    
    ------------------------------------------------------   
    | Left/Right        | X Axis (8)    | Axis 0         |   
    | Up/Down           | Y Axis (8)    | Axis 1         |   
    | Black             | A             | 0              |   
    | Red               | B             | 1              |   
    | Blue              | X             | 2              |   
    | Yellow            | Y             | 3              |   
    | Start             | LB            | 4              |   
    | Coin              | RB (Player 1) | 5              |   
    ------------------------------------------------------

This module / script provides classes for dealing with inputs for the fabarcade.
There are 3 classes Controls,KeyboardControls and JoystickControls

The method names for all the 3 classes are similar and the Keyboard and Joystick
classes are subclasses of Controls class. Allows you to modify inputs by just 
changing the initalising line for the object to the respective class. 
You would not have to touch anything else.

The keyboard controls use the control scheme for the only fabarcade. Please check
our documentation to see which fabarcade input corresponds to what key on the keyboard.

Controls: 
Class for dealing with keyboard and joystick together and combines methods to 
allows you to remain device agnostic.

KeyboardControls: 
Class for dealing with only Keyboard input

JoystickControls:
Class for dealing with only Joystick input

"""


import pygame.key as key
import pygame.event as event
import pygame.joystick as joystick
import pygame
from typing import Any
from Constants import HOME_BUTTON_DOWN,CONTROL_DEBUG

# Set the threshold value for the axis input to check against.
# modify it here to change it everywhere
AXIS_THRESHOLD = 0.97


class Controls:
    """Class for handling controls.
    Combines Keyboard and Joystick calls for the fabarcade. 
    Helps Abstract controls for keyboard and joystick input.
    Returns:
        Controls: Handles both keyboard and joystick input
    """

    def __init__(self) -> None:
        joystick.init()

        # Dictionary for storing joystick objects.
        # key is the instance id and the value is the joystick object
        self.joysticks = [joystick.Joystick(x)
                          for x in range(joystick.get_count())]

        # Helps determine if its P1 or P2
        # I add the joysticks to this array when it is hotplugged,
        # Then when required I check them against this to determine
        # if it is P1 or P2
        self.joystickInstanceIdArray = [
            joy.get_instance_id() for joy in self.joysticks]

        print('Initialised Controls')
        print(self.joysticks)
        print(self.joystickInstanceIdArray)

        #  Implementing chorded input

        # Initialising Key Input
        self._keystate = []

        # Intialise Joystick 1 States
        self._joystick_1_x_axis = 0
        self._joystick_1_y_axis = 0
        self._joystick_1_blue_button = False
        self._joystick_1_yellow_button = False
        self._joystick_1_red_button = False
        self._joystick_1_black_button = False
        self._joystick_1_start_button = False
        self._joystick_1_coin_button = False

        # Initialise Joystick 2 States
        self._joystick_2_x_axis = 0
        self._joystick_2_y_axis = 0
        self._joystick_2_blue_button = False
        self._joystick_2_yellow_button = False
        self._joystick_2_red_button = False
        self._joystick_2_black_button = False
        self._joystick_2_start_button = False
        # Joystick 2 does not have a coin button only take button 5 input from Joystick 1

    """Private exception handling methods
    """

    def _joystick_get(self, idx: int, default: Any) -> joystick.Joystick:
        try:
            return self.joysticks[idx]
        except IndexError:
            return default

    def _check_instance_id(self, instanceId: Any, idx: int) -> bool:
        try:
            return instanceId == self.joystickInstanceIdArray[idx]
        except IndexError:
            return False

    """ Handle Button Events
    """

    def is_button_down(self, e: event.Event):
        eventType = e.type
        return (eventType == pygame.KEYDOWN or eventType == pygame.JOYBUTTONDOWN)

    def is_button_up(self,  e: event.Event):
        eventType = e.type
        return (eventType == pygame.KEYUP or eventType == pygame.JOYBUTTONUP)

    """ Handle Joystick hot plugging  Events
    """

    def add_joystick(self, e: event.Event):
        joy = pygame.joystick.Joystick(e.device_index)
        self.joysticks[joy.get_instance_id()] = joy
        self.joystickInstanceIdArray.append(joy.get_instance_id())
        print(f"Joystick {joy.get_instance_id()} connencted")

    def remove_joyStick(self, e: event.Event):
        del self.joysticks[e.instance_id]
        self.joystickInstanceIdArray.remove(e.instance_id)
        print(f"Joystick {e.instance_id} disconnected")

    """Handle Movement Input
    """

    def is_movement(self, e: event.Event):
        eventType = e.type
        return (eventType == pygame.KEYDOWN or eventType == pygame.JOYAXISMOTION)

    def is_movement_up(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        return ((eventKey == pygame.K_UP or eventKey == pygame.K_w or eventKey == pygame.K_i) or
                (axis == 1 and axisValue <= -AXIS_THRESHOLD))

    def is_movement_down(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        return ((eventKey == pygame.K_DOWN or eventKey == pygame.K_s or eventKey == pygame.K_k) or
                (axis == 1 and axisValue >= AXIS_THRESHOLD))

    def is_movement_left(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        return ((eventKey == pygame.K_LEFT or eventKey == pygame.K_a or eventKey == pygame.K_j) or
                (axis == 0 and axisValue <= -AXIS_THRESHOLD))

    def is_movement_right(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        return ((eventKey == pygame.K_RIGHT or eventKey == pygame.K_d or eventKey == pygame.K_l) or
                (axis == 0 and axisValue >= AXIS_THRESHOLD))

    def is_movement_input(self, e: event.Event):
        return (self.is_movement_up(e) or
                self.is_movement_down(e) or
                self.is_movement_right(e) or
                self.is_movement_left(e))
    """Handle button inputs
    """

    def is_home_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        if (eventKey == pygame.K_ESCAPE or e.type == HOME_BUTTON_DOWN):
            print('isHomeButton activated', eventKey, pygame.K_ESCAPE)
            return True
        else:
            return False

    def is_start_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        return ((eventKey == pygame.K_RETURN or eventKey == pygame.K_b) or
                eventButton == 4)

    # Combined events
    def is_blue_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        return ((eventKey == pygame.K_q or eventKey == pygame.K_t) or
                eventButton == 2)

    def is_yellow_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        return ((eventKey == pygame.K_e or eventKey == pygame.K_u or eventKey == pygame.K_SPACE) or
                eventButton == 3)

    def is_red_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        return ((eventKey == pygame.K_x or eventKey == pygame.K_m) or
                eventButton == 1)

    def is_black_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        return ((eventKey == pygame.K_f or eventKey == pygame.K_n) or
                eventButton == 0)

    def is_coin_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        # This is fine since it is Joystick agnostic
        return (eventKey == pygame.K_c or eventButton == 5)

    # For Player 1

    def is_movement_up_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return ((eventKey == pygame.K_w) or
                (self._check_instance_id(instanceId, 0) and axis == 1 and axisValue <= -AXIS_THRESHOLD))

    def is_movement_down_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return ((eventKey == pygame.K_s) or
                (self._check_instance_id(instanceId, 0) and axis == 1 and axisValue >= AXIS_THRESHOLD))

    def is_movement_left_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return ((eventKey == pygame.K_a) or
                (self._check_instance_id(instanceId, 0) and axis == 0 and axisValue <= -AXIS_THRESHOLD))

    def is_movement_right_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return ((eventKey == pygame.K_d) or
                (self._check_instance_id(instanceId, 0) and axis == 0 and axisValue >= AXIS_THRESHOLD))

    def is_blue_button_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (eventKey == pygame.K_q or
                (self._check_instance_id(instanceId, 0) and eventButton == 2))

    def is_yellow_button_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (eventKey == pygame.K_e or
                (self._check_instance_id(instanceId, 0) and eventButton == 3))

    def is_red_button_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (eventKey == pygame.K_x or
                (self._check_instance_id(instanceId, 0) and eventButton == 1))

    def is_black_button_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (eventKey == pygame.K_f or
                (self._check_instance_id(instanceId, 0) and eventButton == 0))

    # For Player 2
    def is_movement_up_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return ((eventKey == pygame.K_i) or
                (self._check_instance_id(instanceId, 1) and axis == 1 and axisValue <= -AXIS_THRESHOLD))

    def is_movement_down_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return ((eventKey == pygame.K_k) or
                (self._check_instance_id(instanceId, 1) and axis == 1 and axisValue >= AXIS_THRESHOLD))

    def is_movement_left_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return ((eventKey == pygame.K_j) or
                (self._check_instance_id(instanceId, 1) and axis == 0 and axisValue <= -AXIS_THRESHOLD))

    def is_movement_right_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return ((eventKey == pygame.K_l) or
                (self._check_instance_id(instanceId, 1) and axis == 0 and axisValue >= AXIS_THRESHOLD))

    def is_blue_button_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (eventKey == pygame.K_t or (self._check_instance_id(instanceId, 1) and eventButton == 2))

    def is_yellow_button_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (eventKey == pygame.K_u or (self._check_instance_id(instanceId, 1) and eventButton == 3))

    def is_red_button_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (eventKey == pygame.K_m or (self._check_instance_id(instanceId, 1) and eventButton == 1))

    def is_black_button_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (eventKey == pygame.K_n or (self._check_instance_id(instanceId, 1) and eventButton == 0))

    def get_state(self) -> None:
        """Get the current input state. Useful for chording.
        """
        # Initialising Key Input
        self._keystate = key.get_pressed()

        # Intialise Joystick 1 States
        self._joystick_1_x_axis = 0
        self._joystick_1_y_axis = 0
        self._joystick_1_blue_button = False
        self._joystick_1_yellow_button = False
        self._joystick_1_red_button = False
        self._joystick_1_black_button = False
        self._joystick_1_start_button = False
        self._joystick_1_coin_button = False

        # Initialise Joystick 2 States
        self._joystick_2_x_axis = 0
        self._joystick_2_y_axis = 0
        self._joystick_2_blue_button = False
        self._joystick_2_yellow_button = False
        self._joystick_2_red_button = False
        self._joystick_2_black_button = False
        self._joystick_2_start_button = False
        # Joystick 2 does not have a coin button only take button 5 input from Joystick 1

        joy1 = self._joystick_get(0, None)
        joy2 = self._joystick_get(1, None)
        if (joy1 is not None):
            self._joystick_1_x_axis = joy1.get_axis(0)
            self._joystick_1_y_axis = joy1.get_axis(1)
            self._joystick_1_blue_button = joy1.get_button(2)
            self._joystick_1_yellow_button = joy1.get_button(3)
            self._joystick_1_red_button = joy1.get_button(1)
            self._joystick_1_black_button = joy1.get_button(0)
            self._joystick_1_start_button = joy1.get_button(4)
            self._joystick_1_coin_button = joy1.get_button(5)

        if (joy2 is not None):
            self._joystick_2_x_axis = joy2.get_axis(0)
            self._joystick_2_y_axis = joy2.get_axis(1)
            self._joystick_2_blue_button = joy2.get_button(2)
            self._joystick_2_yellow_button = joy2.get_button(3)
            self._joystick_2_red_button = joy2.get_button(1)
            self._joystick_2_black_button = joy2.get_button(0)
            self._joystick_2_start_button = joy2.get_button(4)
            # Joystick 2 does not have a coin button only take button 5 input from Joystick 1
            # self._joystick_2_coin_button = joy2.get_button(5)
        if (CONTROL_DEBUG):
            print(f'joy 1 axis [x,y] = [{self._joystick_1_x_axis},{self._joystick_1_y_axis}] ')
            print(f'joy 1 btns [b,y,r,b] = [{self._joystick_1_blue_button},{self._joystick_1_yellow_button},{self._joystick_1_red_button},{self._joystick_1_black_button}] ')
            print(f'joy 2 axis [x,y] = [{self._joystick_2_x_axis},{self._joystick_2_y_axis}] ')
            print(f'joy 2 btns [b,y,r,b] = [{self._joystick_2_blue_button},{self._joystick_2_yellow_button},{self._joystick_2_red_button},{self._joystick_2_black_button}] ')

        """Combined Methods for both players
        """

    def get_horizontal_axis(self) -> float:
        # Keyboard State
        if (self._keystate[pygame.K_LEFT] or self._keystate[pygame.K_a] or self._keystate[pygame.K_j]):
            return -1.0
        elif (self._keystate[pygame.K_RIGHT] or self._keystate[pygame.K_d] or self._keystate[pygame.K_l]):
            return 1.0
        # Joystick State
        if (self._joystick_1_x_axis > 0.1 or self._joystick_1_x_axis < -0.1):
            return self._joystick_1_x_axis

        if (self._joystick_2_x_axis > 0.1 or self._joystick_2_x_axis < -0.1):
            return self._joystick_2_x_axis

        return 0

    def get_vertical_axis(self) -> float:
        # Keyboard State
        if (self._keystate[pygame.K_UP] or self._keystate[pygame.K_w] or self._keystate[pygame.K_i]):
            return -1.0
        elif (self._keystate[pygame.K_DOWN] or self._keystate[pygame.K_s] or self._keystate[pygame.K_k]):
            return 1.0
        # Joystick State
        if (self._joystick_1_y_axis > 0.1 or self._joystick_1_y_axis < -0.1):
            return self._joystick_1_y_axis

        if (self._joystick_2_y_axis > 0.1 or self._joystick_2_y_axis < -0.1):
            return self._joystick_2_y_axis

        return 0

    def get_blue_button(self) -> bool:
        return ((self._keystate[pygame.K_q] or self._keystate[pygame.K_t]) or
                (self._joystick_1_blue_button or self._joystick_2_blue_button))

    def get_yellow_button(self) -> bool:
        return ((self._keystate[pygame.K_e] or self._keystate[pygame.K_u] or self._keystate[pygame.K_SPACE]) or
                (self._joystick_1_yellow_button or self._joystick_2_yellow_button))

    def get_red_button(self) -> bool:
        return ((self._keystate[pygame.K_x] or self._keystate[pygame.K_m]) or
                (self._joystick_1_red_button or self._joystick_2_red_button))

    def get_black_button(self) -> bool:
        return ((self._keystate[pygame.K_f] or self._keystate[pygame.K_n]) or
                (self._joystick_1_black_button or self._joystick_2_black_button))

    def get_start_button(self) -> bool:
        return ((self._keystate[pygame.K_b] or self._keystate[pygame.K_RETURN]) or
                (self._joystick_1_start_button or self._joystick_2_start_button))

    def get_coin_button(self) -> bool:
        return (self._keystate[pygame.K_c] or self._joystick_1_coin_button)

    """Player 1 State
    """

    def get_horizontal_axis_1(self) -> float:
        # Keyboard State
        if (self._keystate[pygame.K_a]):
            return -1.0
        elif (self._keystate[pygame.K_d]):
            return 1.0
        # Joystick 1 State
        return self._joystick_1_x_axis

    def get_vertical_axis_1(self) -> float:
        # Keyboard State
        if (self._keystate[pygame.K_w]):
            return -1.0
        elif (self._keystate[pygame.K_s]):
            return 1.0
        # Joystick 1 State
        return self._joystick_1_y_axis

    def get_blue_button_1(self) -> bool:
        return (self._keystate[pygame.K_q] or
                self._joystick_1_blue_button)

    def get_yellow_button_1(self) -> bool:
        return (self._keystate[pygame.K_e] or
                self._joystick_1_yellow_button)

    def get_red_button_1(self) -> bool:
        return (self._keystate[pygame.K_x] or
                self._joystick_1_red_button)

    def get_black_button_1(self) -> bool:
        return (self._keystate[pygame.K_f] or
                self._joystick_1_black_button)

    def get_start_button_1(self) -> bool:
        return (self._keystate[pygame.K_b] or
                self._joystick_1_start_button)

    """Player 2 State
    """

    def get_horizontal_axis_2(self) -> float:
        # Keyboard State
        if (self._keystate[pygame.K_j]):
            return -1.0
        elif (self._keystate[pygame.K_l]):
            return 1.0
        # Joystick 2 State
        return self._joystick_2_x_axis

    def get_vertical_axis_2(self) -> float:
        # Keyboard State
        if (self._keystate[pygame.K_i]):
            return -1.0
        elif (self._keystate[pygame.K_k]):
            return 1.0
        # Joystick 2 State
        return self._joystick_2_y_axis

    def get_blue_button_2(self) -> bool:
        return (self._keystate[pygame.K_t] or
                self._joystick_2_blue_button)

    def get_yellow_button_2(self) -> bool:
        return (self._keystate[pygame.K_u] or
                self._joystick_2_yellow_button)

    def get_red_button_2(self) -> bool:
        return (self._keystate[pygame.K_m] or
                self._joystick_2_red_button)

    def get_black_button_2(self) -> bool:
        return (self._keystate[pygame.K_n] or
                self._joystick_2_black_button)

    def get_start_button_2(self) -> bool:
        return (self._keystate[pygame.K_b] or
                self._joystick_2_start_button)


class KeyboardControls(Controls):
    """Class for handling Keyboard Controls.
    For testing in testing environment or during development.
    Implements all the methods for keyboard.

    Args:
        Controls (KeyboardControls): Handles only Keyboard input.
    """

    def __init__(self) -> None:

        print('Initialised Keyboard Controls')
        #  Implementing chorded input

        # Initialising Key Input
        self._keystate = []

    def is_button_down(self, e: event.Event):
        eventType = e.type
        return (eventType == pygame.KEYDOWN)

    def is_button_up(self,  e: event.Event):
        eventType = e.type
        return (eventType == pygame.KEYUP)

    """ Do not Handle Joystick hot plugging  Events
    """

    def add_joystick(self, e: event.Event):
        pass

    def remove_joyStick(self, e: event.Event):
        pass

    """Handle Movement Input
    """

    def is_movement(self, e: event.Event):
        eventType = e.type
        return (eventType == pygame.KEYDOWN)

    def is_movement_up(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_UP or eventKey == pygame.K_w or eventKey == pygame.K_i)

    def is_movement_down(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_DOWN or eventKey == pygame.K_s or eventKey == pygame.K_k)

    def is_movement_left(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_LEFT or eventKey == pygame.K_a or eventKey == pygame.K_j)

    def is_movement_right(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_RIGHT or eventKey == pygame.K_d or eventKey == pygame.K_l)

    def is_movement_input(self, e: event.Event):
        return (self.is_movement_up(e) or
                self.is_movement_down(e) or
                self.is_movement_right(e) or
                self.is_movement_left(e))
    """Handle button inputs
    """

    def is_home_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        if (eventKey == pygame.K_ESCAPE or e.type == HOME_BUTTON_DOWN):
            print('isHomeButton activated', eventKey, pygame.K_ESCAPE)
            return True
        else:
            return False

    def is_start_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_RETURN or eventKey == pygame.K_b)

    # Combined events
    def is_blue_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_q or eventKey == pygame.K_t)

    def is_yellow_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_e or eventKey == pygame.K_u or eventKey == pygame.K_SPACE)

    def is_red_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_x or eventKey == pygame.K_m)

    def is_black_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_f or eventKey == pygame.K_n)

    def is_coin_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_c)

    # For Player 1

    def is_movement_up_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_w)

    def is_movement_down_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_s)

    def is_movement_left_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_a)

    def is_movement_right_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_d)

    def is_blue_button_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_q)

    def is_yellow_button_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_e)

    def is_red_button_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_x)

    def is_black_button_1(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_f)

    # For Player 2
    def is_movement_up_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_i)

    def is_movement_down_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_k)

    def is_movement_left_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_j)

    def is_movement_right_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_l)

    def is_blue_button_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_t)

    def is_yellow_button_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_u)

    def is_red_button_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_m)

    def is_black_button_2(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        return (eventKey == pygame.K_n)

    def get_state(self) -> None:
        """Get the current input state. Useful for chording.
        """
        # Initialising Key Input
        self._keystate = key.get_pressed()

        """Combined Methods for both players
        """

    def get_horizontal_axis(self) -> float:
        # Keyboard State
        if (self._keystate[pygame.K_LEFT] or self._keystate[pygame.K_a] or self._keystate[pygame.K_j]):
            return -1.0
        elif (self._keystate[pygame.K_RIGHT] or self._keystate[pygame.K_d] or self._keystate[pygame.K_l]):
            return 1.0

        return 0

    def get_vertical_axis(self) -> float:
        # Keyboard State
        if (self._keystate[pygame.K_UP] or self._keystate[pygame.K_w] or self._keystate[pygame.K_i]):
            return -1.0
        elif (self._keystate[pygame.K_DOWN] or self._keystate[pygame.K_s] or self._keystate[pygame.K_k]):
            return 1.0

        return 0

    def get_blue_button(self) -> bool:
        return (self._keystate[pygame.K_q] or self._keystate[pygame.K_t])

    def get_yellow_button(self) -> bool:
        return (self._keystate[pygame.K_e] or self._keystate[pygame.K_u] or self._keystate[pygame.K_SPACE])

    def get_red_button(self) -> bool:
        return (self._keystate[pygame.K_x] or self._keystate[pygame.K_m])

    def get_black_button(self) -> bool:
        return (self._keystate[pygame.K_f] or self._keystate[pygame.K_n])

    def get_start_button(self) -> bool:
        return (self._keystate[pygame.K_b] or self._keystate[pygame.K_RETURN])

    def get_coin_button(self) -> bool:
        return (self._keystate[pygame.K_c])

    """Player 1 State
    """

    def get_horizontal_axis_1(self) -> float:
        # Keyboard State
        if (self._keystate[pygame.K_a]):
            return -1.0
        elif (self._keystate[pygame.K_d]):
            return 1.0
        # return idle
        return 0

    def get_vertical_axis_1(self) -> float:
        # Keyboard State
        if (self._keystate[pygame.K_w]):
            return -1.0
        elif (self._keystate[pygame.K_s]):
            return 1.0
        # return idle
        return 0

    def get_blue_button_1(self) -> bool:
        return (self._keystate[pygame.K_q])

    def get_yellow_button_1(self) -> bool:
        return (self._keystate[pygame.K_e])

    def get_red_button_1(self) -> bool:
        return (self._keystate[pygame.K_x])

    def get_black_button_1(self) -> bool:
        return (self._keystate[pygame.K_f])

    def get_start_button_1(self) -> bool:
        return (self._keystate[pygame.K_b])

    """Player 2 State
    """

    def get_horizontal_axis_2(self) -> float:
        # Keyboard State
        if (self._keystate[pygame.K_j]):
            return -1.0
        elif (self._keystate[pygame.K_l]):
            return 1.0
        # return idle
        return 0

    def get_vertical_axis_2(self) -> float:
        # Keyboard State
        if (self._keystate[pygame.K_i]):
            return -1.0
        elif (self._keystate[pygame.K_k]):
            return 1.0
        # return idle
        return 0

    def get_blue_button_2(self) -> bool:
        return (self._keystate[pygame.K_t])

    def get_yellow_button_2(self) -> bool:
        return (self._keystate[pygame.K_u])

    def get_red_button_2(self) -> bool:
        return (self._keystate[pygame.K_m])

    def get_black_button_2(self) -> bool:
        return (self._keystate[pygame.K_n])

    def get_start_button_2(self) -> bool:
        return (self._keystate[pygame.K_b])


class JoystickControls(Controls):
    """Class for handling Joystick Controls.
    For checking Fabarcade inputs with xbox/joystick inputs
    Useful for final code.

    Args:
        Controls (JoystickControls): Handles joystick input only.
    """

    def __init__(self) -> None:
        joystick.init()

        # Dictionary for storing joystick objects.
        # key is the instance id and the value is the joystick object
        self.joysticks = [joystick.Joystick(x)
                          for x in range(joystick.get_count())]

        # Helps determine if its P1 or P2
        # I add the joysticks to this array when it is hotplugged,
        # Then when required I check them against this to determine
        # if it is P1 or P2
        self.joystickInstanceIdArray = [
            joy.get_instance_id() for joy in self.joysticks]

        print('Initialised Joystick Controls')
        print(self.joysticks)
        print(self.joystickInstanceIdArray)

        #  Implementing chorded input

        # Intialise Joystick 1 States
        self._joystick_1_x_axis = 0
        self._joystick_1_y_axis = 0
        self._joystick_1_blue_button = False
        self._joystick_1_yellow_button = False
        self._joystick_1_red_button = False
        self._joystick_1_black_button = False
        self._joystick_1_start_button = False
        self._joystick_1_coin_button = False

        # Initialise Joystick 2 States
        self._joystick_2_x_axis = 0
        self._joystick_2_y_axis = 0
        self._joystick_2_blue_button = False
        self._joystick_2_yellow_button = False
        self._joystick_2_red_button = False
        self._joystick_2_black_button = False
        self._joystick_2_start_button = False
        # Joystick 2 does not have a coin button only take button 5 input from Joystick 1
        # self._joystick_2_coin_button = False

    """Private exception handling methods
    """

    def _joystick_get(self, idx: int, default: Any) -> joystick.Joystick:
        try:
            return self.joysticks[idx]
        except IndexError:
            return default

    def _check_instance_id(self, instanceId: Any, idx: int) -> bool:
        try:
            return instanceId == self.joystickInstanceIdArray[idx]
        except IndexError:
            return False

    """ Handle Button Events
    """

    def is_button_down(self, e: event.Event):
        eventType = e.type
        return (eventType == pygame.JOYBUTTONDOWN)

    def is_button_up(self,  e: event.Event):
        eventType = e.type
        return (eventType == pygame.JOYBUTTONUP)

    """ Handle Joystick hot plugging  Events
    """

    def add_joystick(self, e: event.Event):
        joy = pygame.joystick.Joystick(e.device_index)
        self.joysticks[joy.get_instance_id()] = joy
        self.joystickInstanceIdArray.append(joy.get_instance_id())
        print(f"Joystick {joy.get_instance_id()} connencted")

    def remove_joyStick(self, e: event.Event):
        del self.joysticks[e.instance_id]
        self.joystickInstanceIdArray.remove(e.instance_id)
        print(f"Joystick {e.instance_id} disconnected")

    """Handle Movement Input
    """

    def is_movement(self, e: event.Event):
        eventType = e.type
        return (eventType == pygame.JOYAXISMOTION)

    def is_movement_up(self, e: event.Event):
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        return (axis == 1 and axisValue <= -AXIS_THRESHOLD)

    def is_movement_down(self, e: event.Event):
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        return (axis == 1 and axisValue >= AXIS_THRESHOLD)

    def is_movement_left(self, e: event.Event):
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        return (axis == 0 and axisValue <= -AXIS_THRESHOLD)

    def is_movement_right(self, e: event.Event):
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        return (axis == 0 and axisValue >= AXIS_THRESHOLD)

    def is_movement_input(self, e: event.Event):
        return (self.is_movement_up(e) or
                self.is_movement_down(e) or
                self.is_movement_right(e) or
                self.is_movement_left(e))
    """Handle button inputs
    """

    def is_home_button(self, e: event.Event):
        eventKey = e.key if e.type == pygame.KEYDOWN else None
        if (eventKey == pygame.K_ESCAPE or e.type == HOME_BUTTON_DOWN):
            print('isHomeButton activated', eventKey, pygame.K_ESCAPE)
            return True
        else:
            return False

    def is_start_button(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        return (eventButton == 4)

    # Combined events
    def is_blue_button(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        return (eventButton == 2)

    def is_yellow_button(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        return (eventButton == 3)

    def is_red_button(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        return (eventButton == 1)

    def is_black_button(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        return (eventButton == 0)

    def is_coin_button(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        # Joystick Agnostic so this should be fine
        return (eventButton == 5)

    # For Player 1

    def is_movement_up_1(self, e: event.Event):
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return (self._check_instance_id(instanceId, 0) and axis == 1 and axisValue <= -AXIS_THRESHOLD)

    def is_movement_down_1(self, e: event.Event):
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return (self._check_instance_id(instanceId, 0) and axis == 1 and axisValue >= AXIS_THRESHOLD)

    def is_movement_left_1(self, e: event.Event):
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return (self._check_instance_id(instanceId, 0) and axis == 0 and axisValue <= -AXIS_THRESHOLD)

    def is_movement_right_1(self, e: event.Event):
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return (self._check_instance_id(instanceId, 0) and axis == 0 and axisValue >= AXIS_THRESHOLD)

    def is_blue_button_1(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (self._check_instance_id(instanceId, 0) and eventButton == 2)

    def is_yellow_button_1(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (self._check_instance_id(instanceId, 0) and eventButton == 3)

    def is_red_button_1(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (self._check_instance_id(instanceId, 0) and eventButton == 1)

    def is_black_button_1(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (self._check_instance_id(instanceId, 0) and eventButton == 0)

    # For Player 2
    def is_movement_up_2(self, e: event.Event):
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return (self._check_instance_id(instanceId, 1) and axis == 1 and axisValue <= -AXIS_THRESHOLD)

    def is_movement_down_2(self, e: event.Event):
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return (self._check_instance_id(instanceId, 1) and axis == 1 and axisValue >= AXIS_THRESHOLD)

    def is_movement_left_2(self, e: event.Event):
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return (self._check_instance_id(instanceId, 1) and axis == 0 and axisValue <= -AXIS_THRESHOLD)

    def is_movement_right_2(self, e: event.Event):
        axis = e.axis if e.type == pygame.JOYAXISMOTION else None
        axisValue = e.value if e.type == pygame.JOYAXISMOTION else 0
        instanceId = e.instance_id if e.type == pygame.JOYAXISMOTION else None

        return (self._check_instance_id(instanceId, 1) and axis == 0 and axisValue >= AXIS_THRESHOLD)

    def is_blue_button_2(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (self._check_instance_id(instanceId, 1) and eventButton == 2)

    def is_yellow_button_2(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (self._check_instance_id(instanceId, 1) and eventButton == 3)

    def is_red_button_2(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (self._check_instance_id(instanceId, 1) and eventButton == 1)

    def is_black_button_2(self, e: event.Event):
        eventButton = e.button if e.type == pygame.JOYBUTTONDOWN else None
        instanceId = e.instance_id if e.type == pygame.JOYBUTTONDOWN else None

        return (self._check_instance_id(instanceId, 1) and eventButton == 0)

    def get_state(self) -> None:
        """Get the current input state. Useful for chording.
        """
        # Intialise Joystick 1 States
        self._joystick_1_x_axis = 0
        self._joystick_1_y_axis = 0
        self._joystick_1_blue_button = False
        self._joystick_1_yellow_button = False
        self._joystick_1_red_button = False
        self._joystick_1_black_button = False
        self._joystick_1_start_button = False
        self._joystick_1_coin_button = False

        # Initialise Joystick 2 States
        self._joystick_2_x_axis = 0
        self._joystick_2_y_axis = 0
        self._joystick_2_blue_button = False
        self._joystick_2_yellow_button = False
        self._joystick_2_red_button = False
        self._joystick_2_black_button = False
        self._joystick_2_start_button = False
        # Joystick 2 does not have a coin button only take button 5 input from Joystick 1
        # self._joystick_2_coin_button = False

        joy1 = self._joystick_get(0, None)
        joy2 = self._joystick_get(1, None)
        if (joy1 is not None):
            self._joystick_1_x_axis = joy1.get_axis(0)
            self._joystick_1_y_axis = joy1.get_axis(1)
            self._joystick_1_blue_button = joy1.get_button(2)
            self._joystick_1_yellow_button = joy1.get_button(3)
            self._joystick_1_red_button = joy1.get_button(1)
            self._joystick_1_black_button = joy1.get_button(0)
            self._joystick_1_start_button = joy1.get_button(4)
            self._joystick_1_coin_button = joy1.get_button(5)

        if (joy2 is not None):
            self._joystick_2_x_axis = joy2.get_axis(0)
            self._joystick_2_y_axis = joy2.get_axis(1)
            self._joystick_2_blue_button = joy2.get_button(2)
            self._joystick_2_yellow_button = joy2.get_button(3)
            self._joystick_2_red_button = joy2.get_button(1)
            self._joystick_2_black_button = joy2.get_button(0)
            self._joystick_2_start_button = joy2.get_button(4)
            # Joystick 2 does not have a coin button only take button 5 input from Joystick 1
            # self._joystick_2_coin_button = joy2.get_button(5)
    
        if (CONTROL_DEBUG):
            print(f'joy 1 axis [x,y] = [{self._joystick_1_x_axis},{self._joystick_1_y_axis}] ')
            print(f'joy 1 btns [b,y,r,b] = [{self._joystick_1_blue_button},{self._joystick_1_yellow_button},{self._joystick_1_red_button},{self._joystick_1_black_button}] ')
            print(f'joy 2 axis [x,y] = [{self._joystick_2_x_axis},{self._joystick_2_y_axis}] ')
            print(f'joy 2 btns [b,y,r,b] = [{self._joystick_2_blue_button},{self._joystick_2_yellow_button},{self._joystick_2_red_button},{self._joystick_2_black_button}] ')


        """Combined Methods for both players
        """

    def get_horizontal_axis(self) -> float:
        # Joystick State
        if (self._joystick_1_x_axis > 0.1 or self._joystick_1_x_axis < -0.1):
            return self._joystick_1_x_axis

        if (self._joystick_2_x_axis > 0.1 or self._joystick_2_x_axis < -0.1):
            return self._joystick_2_x_axis

        return 0

    def get_vertical_axis(self) -> float:
        # Joystick State
        if (self._joystick_1_y_axis > 0.1 or self._joystick_1_y_axis < -0.1):
            return self._joystick_1_y_axis

        if (self._joystick_2_y_axis > 0.1 or self._joystick_2_y_axis < -0.1):
            return self._joystick_2_y_axis

        return 0

    def get_blue_button(self) -> bool:
        return (self._joystick_1_blue_button or self._joystick_2_blue_button)

    def get_yellow_button(self) -> bool:
        return (self._joystick_1_yellow_button or self._joystick_2_yellow_button)

    def get_red_button(self) -> bool:
        return (self._joystick_1_red_button or self._joystick_2_red_button)

    def get_black_button(self) -> bool:
        return (self._joystick_1_black_button or self._joystick_2_black_button)

    def get_start_button(self) -> bool:
        return (self._joystick_1_start_button or self._joystick_2_start_button)

    def get_coin_button(self) -> bool:
        return (self._joystick_1_coin_button)

    """Player 1 State
    """

    def get_horizontal_axis_1(self) -> float:
        # Joystick 1 State
        return self._joystick_1_x_axis

    def get_vertical_axis_1(self) -> float:
        # Joystick 1 State
        return self._joystick_1_y_axis

    def get_blue_button_1(self) -> bool:
        return (self._joystick_1_blue_button)

    def get_yellow_button_1(self) -> bool:
        return (self._joystick_1_yellow_button)

    def get_red_button_1(self) -> bool:
        return (self._joystick_1_red_button)

    def get_black_button_1(self) -> bool:
        return (self._joystick_1_black_button)

    def get_start_button_1(self) -> bool:
        return (self._joystick_1_start_button)

    """Player 2 State
    """

    def get_horizontal_axis_2(self) -> float:
        # Joystick 2 State
        return self._joystick_2_x_axis

    def get_vertical_axis_2(self) -> float:
        # Joystick 2 State
        return self._joystick_2_y_axis

    def get_blue_button_2(self) -> bool:
        return (self._joystick_2_blue_button)

    def get_yellow_button_2(self) -> bool:
        return (self._joystick_2_yellow_button)

    def get_red_button_2(self) -> bool:
        return (self._joystick_2_red_button)

    def get_black_button_2(self) -> bool:
        return (self._joystick_2_black_button)

    def get_start_button_2(self) -> bool:
        return (self._joystick_2_start_button)
