import pygame
import Colors
from Controls import Controls
from pygame.surface import Surface
from typing import List
from pygame.locals import *
from Constants import FONT_FILE_PATH
from Manifest import Manifest
from Description import *


class Grid():
    """ The given elements are displayed in a grid of 12 squares per page
        A cursor indicates the active element
        Next to the Grid a box with a text description of the active element is displayed
        There can be multiple pages
    """

    def __init__(self, screen: Surface, width: float, height: float, manifestList:List[Manifest], id:str,onChange=None,onChanged=None):
        """Initialise the Grid

        Args:
            screen (Surface): The surface to display the grid on
            width (float): Width of the grid
            height (float): Height of the grid
            manifestList (List): List of elements to display in the grid
            identifier (str): String name for identifying the grid
        """


        self.screen = screen
        self.width = width
        self.height = height
        self.manifestList = manifestList
        self.id = id


        self.surface = (Surface((width, height)))
        self.size = len(manifestList)
        self.activeIndex = 0
        self.page = 1
        self.numberOfPages = 1+ self.size//12
        
        # Set the callbacks for the events
        self.onChange = onChange
        self.onChanged = onChanged

        
        
        # Set the sizes for the grid
        self.slice_x= width/12
        self.slice_y = height/12
        
        self.square = self.slice_x * 1.4
        self.ix = self.slice_x * 1
        self.iy = self.slice_x * 0.5
        self.border = 4

        # Initialise the images in the grid
        self.grid_images_list = []
        for manifest in self.manifestList:
            image = pygame.transform.smoothscale(pygame.image.load(manifest.image).convert(),(self.square-self.border*2,self.square-self.border*2))
            self.grid_images_list.append(image)


        self.font = pygame.font.Font(FONT_FILE_PATH, 10)
        self.fontTitle = pygame.font.Font(FONT_FILE_PATH, 12)

        # Define a rectangle around the squares in the grid
        self.grid_rect = Rect(self.ix,self.iy,4*self.square + 15,3*self.square + 8)
        # Initialise the descrpition
        self.description = Description(self.surface, self.slice_x*7.5, self.slice_y*2, self.slice_y*7, self.grid_rect.height-10, self.grid_rect)
        self.description.infoSurfaceRect.right = self.width - self.ix

        self.draw()

    def getActiveIndex(self):
        return self.activeIndex

    def setActiveIndex(self,newIndex):
        self.activeIndex = newIndex
        self.page = self.activeIndex // 12 + 1

        self.draw()

    def handleOnChange(self):
        if self.onChange is not None:
            self.onChange(self)

    def handleOnChanged(self):
        if self.onChanged is not None:
            self.onChanged(self)

    # draw the grid view as it will be shown on screen
    def draw(self):
        # Initialise the Surfaces and rects of the Grid Items
        self.surfaceList: List[Surface] = []
        self.imageList: List[Rect] = []

        start_index = (self.page - 1)*12
        end_index = (self.page)*12
        for x in range(start_index,end_index):
            surface = Surface((self.square, self.square))
            self.surfaceList.append(surface)
            image = pygame.draw.rect(surface, Colors.GRAY,(self.border,self.border,self.square-self.border*2,self.square-self.border*2))
            if x < self.size:
                surface.blit(self.grid_images_list[x],(self.border,self.border,self.square-self.border,self.square-self.border))
            
            image.x = self.ix + (x%4)*(self.square+self.border*2)
            image.y = self.iy + ((x-start_index)//4)*(self.square+self.border*2)
            self.imageList.append(image)

        # Initialise cursor
        self.cursor = pygame.draw.rect(self.surfaceList[self.activeIndex%12], Colors.WHITE,(0,0,self.square,self.square), width=self.border) 

        # Draw the game title and page index 
        gridItemName = self.manifestList[self.activeIndex].name  + ' (' + self.manifestList[self.activeIndex].year + ')'         
        font_title = pygame.font.Font(FONT_FILE_PATH,20)
        self.gameTitle = font_title.render(gridItemName,True,Colors.WHITE)
        self.gameTitleRect = Rect(self.surface.get_rect().centerx - (self.gameTitle.get_rect().width/2), self.slice_y * 11, self.gameTitle.get_width(),20)
        self.gameTitleRect.centerx = ((self.grid_rect.right - self.grid_rect.left)/2) + self.grid_rect.left
        font_page = pygame.font.Font(FONT_FILE_PATH,12)
        self.pageIndex = font_page.render(str(self.page) + "/" + str(self.numberOfPages), True, Colors.WHITE)
        self.pageIndexRect = Rect(self.slice_x * 10,0,self.pageIndex.get_width(),10)
        self.pageIndexRect.bottom = self.gameTitleRect.bottom


        self.description.render(self.manifestList[self.activeIndex].name, self.manifestList[self.activeIndex].description,self.manifestList[self.activeIndex].group_members)
        
        self.pageIndexRect.right = self.description.infoSurfaceRect.right
        
    # display the grid view on screen
    def render(self):
        self.surface.fill(Colors.BLACK)
        # display grid items
        n = min(12, self.size - (12 * (self.page-1)))
        for x in range(n):
            self.surface.blit(self.surfaceList[x],self.imageList[x])

        self.surface.blit(self.gameTitle,self.gameTitleRect)
        self.surface.blit(self.pageIndex,self.pageIndexRect)
        
        # display description
        self.description.render(self.manifestList[self.activeIndex].name, self.manifestList[self.activeIndex].description, self.manifestList[self.activeIndex].group_members)

        # blit grid surface on screen
        self.screen.blit(self.surface, self.surface.get_rect())

    # methods for cursor movement in each direction
    def cursorLeft(self):
        if(self.activeIndex%4 != 0):
            self.activeIndex -= 1
        else:
            self.activeIndex -= 9
            self.page -= 1
            if(self.page < 1):
                self.page = self.numberOfPages
                self.activeIndex = min(self.activeIndex+self.numberOfPages*12, self.size-1)
        self.handleOnChange()
        self.draw()
        # Done changing the index
        self.handleOnChanged()

    def cursorRight(self):
        if(self.activeIndex%4 != 3):
            self.activeIndex += 1
            if(self.activeIndex >= self.size):
                self.activeIndex = 0
                self.page = 1
        else:
            self.activeIndex += 9
            self.page += 1
            if(self.page > self.numberOfPages):
                self.page = 1
                self.activeIndex -= self.numberOfPages*12
            else:
                while(self.activeIndex >= self.size):
                    self.activeIndex -= 4
        self.handleOnChange()
        self.draw()
        # Done changing the index
        self.handleOnChanged()

    def cursorUp(self):
        print(str(self.activeIndex))
        self.activeIndex -= 4
        print(str(self.activeIndex))
        if(self.activeIndex%12 > 7):
            print(str(self.activeIndex))
            self.activeIndex += 12
            print(str(self.activeIndex))
            while(self.activeIndex >= self.size):
                print(str(self.activeIndex))
                self.activeIndex -= 4
                print(str(self.activeIndex))
        self.handleOnChange()
        self.draw()
        # Done changing the index
        self.handleOnChanged()

    def cursorDown(self):
        self.activeIndex += 4
        if(self.activeIndex%12 < 4):
            self.activeIndex -= 12
        if(self.activeIndex >= self.size):
            while(self.activeIndex%12 >= 4):
                self.activeIndex -= 4
        self.handleOnChange()
        self.draw()
        # Done changing the index
        self.handleOnChanged()

    # move the cursor according to the input
    def moveCursor(self, event:pygame.event.Event, controls:Controls):
        if controls.is_movement_left(event):
            self.cursorLeft()
        elif controls.is_movement_right(event):
            self.cursorRight()
        elif controls.is_movement_up(event):
            self.cursorUp()
        elif controls.is_movement_down(event):
            self.cursorDown()    

    # start the currently selected game
    def onSelect(self, callback=None, cargs=()):
        print("Grid onSelect")
        if(callback is not None):
            callback(cargs, self.manifestList[self.activeIndex].name)



