import asyncio
from subprocess import Popen
from queue import Queue
from typing import List
import pygame
from Controls import Controls,KeyboardControls,JoystickControls,AXIS_THRESHOLD
try:
    from evdev import InputDevice, categorize, ecodes,list_devices,UInput
except Exception:
    print('Evdev not found')
from traceback import print_exception
from ProcessHandler import handle_terminate_process
# from selectors import DefaultSelector,EVENT_READ

class LegacyControls:
    """
    This class was written to emulate the keyboard controls of the first fabarcade. 
    It does this by directly reading the input event stream from the OS for the 
    joystick input and creating a virtual keyboard and writing to that keyboard to 
    simulate and generate keypresses. The evdev library is used  to read and write
    to the eventstream.
    This class only works directly on the pi.
    """
    def __init__(self,controls:Controls) -> None:
        self.controls = controls
        try:
            # Translation for the joystick keys to Fabarcade keys
            self.joystickMap = {
                0:{
                    'left':ecodes.KEY_A,
                    'right':ecodes.KEY_D,
                    'up':ecodes.KEY_W,
                    'down':ecodes.KEY_S,
                    'start_button': ecodes.KEY_1,
                    'coin_button': ecodes.KEY_C,
                    'blue_button': ecodes.KEY_Q,
                    'yellow_button': ecodes.KEY_E,
                    'red_button': ecodes.KEY_X,
                    'black_button': ecodes.KEY_F,
                },
                1:{
                    'left':ecodes.KEY_J,
                    'right':ecodes.KEY_L,
                    'up':ecodes.KEY_I,
                    'down':ecodes.KEY_K,
                    'start_button': ecodes.KEY_2,
                    'coin_button': ecodes.KEY_C,
                    'blue_button': ecodes.KEY_T,
                    'yellow_button': ecodes.KEY_U,
                    'red_button': ecodes.KEY_M,
                    'black_button': ecodes.KEY_N,
                }
            }
        
            self._keystate = { }
            
            self._devices = [InputDevice(path) for path in list_devices()]
            self._joysticks = []
            self._joyNames = [joy.get_name() for joy in controls.joysticks]
            # Search for the joystick devices and add them to _joysticks list
            for device in self._devices:
                print(device.path,device.name,device.phys)
                for i,joyName in enumerate(self._joyNames):
                    if device.name == joyName:
                        self._joysticks.append(device.path)
                        self._joyNames.pop(i)
                        print('got a joystick')
                        break
                    device.close()
            # Async io event loop
            self.loop = asyncio.get_event_loop()
            
            # Keep track of all the tasks in the event loop for exit.
            self.tasks_list = []
            # self.dev = None
        
            # Initalise a virtual UInput device.
            # This can only be done on a linux system. 
            # This code will crash in windows.
            self.dev = UInput()

            print('Successfully, Initialised Legacy Controls...')
        except Exception as error:
            print(error)
            # Gracefully Deinitialise legacy mode if there is an error
            # If ran on an OS other than linux or on the pi
            self.dev = None
            print('Legacy Controls was not able to initalise. /etc/uinput was not found or some other exception.')
            print('Proceeding ahead without legacy layer.')
        
    

    
    def checkButton(self,button_state:bool,key_code):
        if button_state:
            self._keystate[key_code] = True
            self.dev.write(ecodes.EV_KEY,key_code,1)
        elif self._keystate.get(key_code) == True: 
            self._keystate[key_code] = False
            self.dev.write(ecodes.EV_KEY,key_code,0)
        self.dev.syn()
            
    def checkAxis(self,axis_value,key_code_neg,key_code_pos):
        if axis_value> AXIS_THRESHOLD:
            self._keystate[key_code_pos] = True
            self.dev.write(ecodes.EV_KEY,key_code_pos,1)
        elif axis_value< -AXIS_THRESHOLD:
            self._keystate[key_code_neg] = True
            self.dev.write(ecodes.EV_KEY,key_code_neg,1)
            
        
        if abs(axis_value) < 0.1 and self._keystate.get(key_code_neg) == True:
            self._keystate[key_code_neg] = False
            self.dev.write(ecodes.EV_KEY,key_code_neg,0)


        if abs(axis_value) < 0.1 and self._keystate.get(key_code_pos) == True:
            self._keystate[key_code_pos] = False
            self.dev.write(ecodes.EV_KEY,key_code_pos,0)
        
        self.dev.syn()

    async def emulate_device(self,devicePath:str,deviceIndex:int):
        """Emulate a device for the previous fabarcade

        Args:
            devicePath (str): Device path to the event stream
            deviceIndex (int): Device index for the right input translation for the joystick
        """
        # Open device
        print('open device')
        device = InputDevice(devicePath)
        try:
            async for event in device.async_read_loop():
                print('new iteration')
                print(device, event)
                print(categorize(event))
                if event.type == ecodes.EV_ABS:
                    axisVal = event.value/32767
                    axisVal -= 1
                    print(axisVal)
                    if event.code == ecodes.ABS_X:
                        self.checkAxis(axisVal,self.joystickMap[deviceIndex]['left'],self.joystickMap[deviceIndex]['right'])
                    elif event.code == ecodes.ABS_Y:
                        self.checkAxis(axisVal,self.joystickMap[deviceIndex]['up'],self.joystickMap[deviceIndex]['down'])
                elif event.type == ecodes.EV_KEY:
                    btnState = event.value == 1
                    if event.code == ecodes.BTN_WEST:
                        # Start Button
                        self.checkButton(btnState,self.joystickMap[deviceIndex]['start_button'])
                    elif event.code == ecodes.BTN_Z:
                        # Coin button
                        self.checkButton(btnState,self.joystickMap[deviceIndex]['coin_button'])
                    elif event.code == ecodes.BTN_C:
                        # Blue button
                        self.checkButton(btnState,self.joystickMap[deviceIndex]['blue_button'])
                    elif event.code == ecodes.BTN_NORTH:
                        # Yellow Button
                        self.checkButton(btnState,self.joystickMap[deviceIndex]['yellow_button'])
                    elif event.code == ecodes.BTN_EAST:
                        # Red button
                        self.checkButton(btnState,self.joystickMap[deviceIndex]['red_button'])
                    elif event.code == ecodes.BTN_SOUTH:
                        # Black Button
                        self.checkButton(btnState,self.joystickMap[deviceIndex]['black_button'])
        except asyncio.CancelledError:
            print('task is cancelled')        
        finally:
            print('close device')
            device.close()
    
    async def exit_emulation(self,Loop,callback):
        """Background loop that listens for home button input for exit.
        Args:
            process_queue (Queue[Popen]): Process Queue
            Loop (_type_): Loop Variable keeping track of the loop state.
            callback (function): Callback function to execute at the end.
        """
        while Loop.isLegacy():
            await asyncio.sleep(0.5)
            print('in legacy loop')
            for event in pygame.event.get():
                if self.controls.is_home_button(event):
                    handle_terminate_process()
                    pygame.event.set_allowed(None)
                    tasks = asyncio.all_tasks()
                    for t in tasks:
                        t.cancel()
                    self.tasks_list = tasks
                    self.loop.stop()
                    callback()

    async def emulate(self,Loop,callback,event_loop:asyncio.AbstractEventLoop):
        """Start emulation. Will only run on linux after initialising the evdev module.

        Args:
            Loop (_type_): Loop state variable
            callback (function): callback to call once it needs to be exited.
            event_loop (asyncio.AbstractEventLoop): Asyncio event loop.
            process_queue (Queue[Popen]): Process Queue keeping track of the running task
        """
        # Set the current async event loop to itself. To maintain a reference to it. 
        self.loop = event_loop
        if (self.dev is not None):
            # Start the background task to allow exit from the application.
            exit_task = asyncio.create_task(self.exit_emulation(Loop,callback))            
            
            # For one joystick when testing
            if len(self._joysticks)== 1:
                for id,path in enumerate(self._joysticks):
                    task =  asyncio.create_task(self.emulate_device(path,id))
                    # self.tasks_list.append(task)
                    await task
            elif len(self._joysticks) >= 2:
                # This is when on the pi it detects both joysticks. 
                # This was done this way since asyncio.gather does not accept
                # not accept a list.
                for id,path in enumerate(self._joysticks):
                    print(id, path)

                path1 = self._joysticks[0]
                path2 = self._joysticks[1]
                print(path1,path2)
                print('started both tasks with gather')
                try:
                    task1 = asyncio.create_task(self.emulate_device(path1,0))
                    task2 = asyncio.create_task(self.emulate_device(path2,1))
                    await asyncio.gather(task1,task2)
                except asyncio.CancelledError:
                    print('gather is cancelled.')
                
            
